//
//  SPK_CalligraphyLineStructure.m
//  SPK-Calligraphy
//
//  Created by Toby Harris on 25/03/2009.
//  Copyright (c) 2009 *spark audiovisual. All rights reserved.
//
//  'SPK-Calligraphy' is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  'SPK-Calligraphy' is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

/* It's highly recommended to use CGL macros instead of changing the current context for plug-ins that perform OpenGL rendering */
#import <OpenGL/CGLMacro.h>

#import "SPK_CalligraphyLineStructure.h"

#define	kQCPlugIn_Name				@"SPK Calligraphy - Line Structure"
#define	kQCPlugIn_Description		@"Will draw a line as if by a chisel nib at 45 degrees. Line Structure is a list of co-ordinates and corresponding pressure, ie. a structure of x,y,p structures, all accesses by index not key. x,y and width are in qc units and p should be normalised. This patch is somewhat interchangeable with Kineme GL Line Structure."

@implementation SPK_CalligraphyLineStructure

/*
Here you need to declare the input / output properties as dynamic as Quartz Composer will handle their implementation
@dynamic inputFoo, outputBar;
*/

@dynamic inputLineStructure, inputColor, inputFlowDecay, inputFlowMinimum, inputWidth;

+ (NSDictionary*) attributes
{
	/*
	Return a dictionary of attributes describing the plug-in (QCPlugInAttributeNameKey, QCPlugInAttributeDescriptionKey...).
	*/
	
	return [NSDictionary dictionaryWithObjectsAndKeys:kQCPlugIn_Name, QCPlugInAttributeNameKey, kQCPlugIn_Description, QCPlugInAttributeDescriptionKey, nil];
}

+ (NSDictionary*) attributesForPropertyPortWithKey:(NSString*)key
{
	/*
	Specify the optional attributes for property based ports (QCPortAttributeNameKey, QCPortAttributeDefaultValueKey...).
	*/
	
	if([key isEqualToString:@"inputLineStructure"])
		return [NSDictionary dictionaryWithObjectsAndKeys:@"Line Structure", QCPortAttributeNameKey, nil];
	if([key isEqualToString:@"inputColor"])
		return [NSDictionary dictionaryWithObjectsAndKeys:@"Line Colour", QCPortAttributeNameKey, nil];
	if([key isEqualToString:@"inputWidth"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"Line Width", QCPortAttributeNameKey,
				[NSNumber numberWithDouble:0.05], QCPortAttributeDefaultValueKey,
				[NSNumber numberWithDouble:0.001], QCPortAttributeMinimumValueKey,
				nil];
	if([key isEqualToString:@"inputFlowDecay"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"Flow Decay Proportion", QCPortAttributeNameKey,
				[NSNumber numberWithDouble:0.9], QCPortAttributeDefaultValueKey,
				[NSNumber numberWithDouble:0], QCPortAttributeMinimumValueKey,
				[NSNumber numberWithDouble:1], QCPortAttributeMaximumValueKey,
				nil];
	if([key isEqualToString:@"inputFlowMinimum"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"Flow Minimum Proportion", QCPortAttributeNameKey,
				[NSNumber numberWithDouble:0.2], QCPortAttributeDefaultValueKey,
				[NSNumber numberWithDouble:0], QCPortAttributeMinimumValueKey,
				[NSNumber numberWithDouble:1], QCPortAttributeMaximumValueKey,
				nil];
	return nil;
}

+ (QCPlugInExecutionMode) executionMode
{
	/*
	Return the execution mode of the plug-in: kQCPlugInExecutionModeProvider, kQCPlugInExecutionModeProcessor, or kQCPlugInExecutionModeConsumer.
	*/
	
	return kQCPlugInExecutionModeConsumer;
}

+ (QCPlugInTimeMode) timeMode
{
	/*
	Return the time dependency mode of the plug-in: kQCPlugInTimeModeNone, kQCPlugInTimeModeIdle or kQCPlugInTimeModeTimeBase.
	*/
	
	return kQCPlugInTimeModeNone;
}

- (id) init
{
	if(self = [super init]) {
		/*
		Allocate any permanent resource required by the plug-in.
		*/
	}
	
	return self;
}

- (void) finalize
{
	/*
	Release any non garbage collected resources created in -init.
	*/
	
	[super finalize];
}

- (void) dealloc
{
	/*
	Release any resources created in -init.
	*/
	
	[super dealloc];
}

+ (NSArray*) plugInKeys
{
	/*
	Return a list of the KVC keys corresponding to the internal settings of the plug-in.
	*/
	
	return nil;
}

- (id) serializedValueForKey:(NSString*)key;
{
	/*
	Provide custom serialization for the plug-in internal settings that are not values complying to the <NSCoding> protocol.
	The return object must be nil or a PList compatible i.e. NSString, NSNumber, NSDate, NSData, NSArray or NSDictionary.
	*/
	
	return [super serializedValueForKey:key];
}

- (void) setSerializedValue:(id)serializedValue forKey:(NSString*)key
{
	/*
	Provide deserialization for the plug-in internal settings that were custom serialized in -serializedValueForKey.
	Deserialize the value, then call [self setValue:value forKey:key] to set the corresponding internal setting of the plug-in instance to that deserialized value.
	*/
	
	[super setSerializedValue:serializedValue forKey:key];
}

- (QCPlugInViewController*) createViewController
{
	/*
	Return a new QCPlugInViewController to edit the internal settings of this plug-in instance.
	You can return a subclass of QCPlugInViewController if necessary.
	*/
	
	return [[QCPlugInViewController alloc] initWithPlugIn:self viewNibName:@"Settings"];
}

@end

@implementation SPK_CalligraphyLineStructure (Execution)

- (BOOL) startExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition starts: perform any required setup for the plug-in.
	Return NO in case of fatal failure (this will prevent rendering of the composition to start).
	*/
	
	return YES;
}

- (void) enableExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance starts being used by Quartz Composer.
	*/
}

- (BOOL) execute:(id<QCPlugInContext>)context atTime:(NSTimeInterval)time withArguments:(NSDictionary*)arguments
{
	CGLContextObj					cgl_ctx = [context CGLContextObj];
	GLenum							error;
	GLint							blendEnabled;
	
	const CGFloat*					colorComponents;
	float							decay, baseAlpha, runningAlpha;
	NSArray*						xypCoordinates;
	double							x, y, width, offset;
	
	if(cgl_ctx == NULL)
		return NO;
	
	//// TASK: Set up blend mode to allow alpha
	blendEnabled = glIsEnabled(GL_BLEND);
	if(!blendEnabled)
		glEnable(GL_BLEND);
	glBlendEquation(GL_FUNC_ADD);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	// cache inputs we're going to need repeatedly
	width = self.inputWidth;
	decay = self.inputFlowDecay;
		
	//// TASK: Render the line as a QUAD STRIP as if we had a chisel nib at 45
	glBegin(GL_QUAD_STRIP);
	
	// get RGBA components (the CGColorRef is guaranteed to be of type RGBA, no need to save / restore as the current color is part of the GL_CURRENT_BIT)
	colorComponents = CGColorGetComponents(self.inputColor);	

	// set our flow fade out vars
	baseAlpha = colorComponents[3] * self.inputFlowMinimum;
	runningAlpha = colorComponents[3] - baseAlpha;
	
	// plot the coordinates
	for (xypCoordinates in self.inputLineStructure) {
		x = [[xypCoordinates objectAtIndex:0] doubleValue];
		y = [[xypCoordinates objectAtIndex:1] doubleValue];
		offset = (width / 2) * ([[xypCoordinates objectAtIndex:2] doubleValue] + 0.5);
		runningAlpha *= decay;
		
		glColor4f(colorComponents[0], colorComponents[1], colorComponents[2], baseAlpha + runningAlpha);
		glVertex2d(x+offset, y+offset);
		glVertex2d(x-offset, y-offset);
	}
	
	glEnd();
	
	if(!blendEnabled)
		glDisable(GL_BLEND);
	
	// check for OpenGL errors
	if(error = glGetError())
		[context logMessage:@"OpenGL error %04X", error];
	
	return (error ? NO : YES);
}

- (void) disableExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance stops being used by Quartz Composer.
	*/
}

- (void) stopExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition stops: perform any required cleanup for the plug-in.
	*/
}

@end
